### The MATLAB module does not work with python 3.7.0. Use 3.6.0 release or older.

import sys
import time
from collections import defaultdict

import numpy as np

from holisticai.utils.transformers.bias import SensitiveGroups

EPSILON = 0.0001


class TreeNode:
    def __init__(self):
        self.children = []

    def set_cluster(self, cluster):
        self.cluster = cluster

    def add_child(self, child):
        self.children.append(child)

    def populate_colors(self, colors):
        "Populate auxiliary lists of red and blue points for each node, bottom-up"
        self.reds = []
        self.blues = []
        if len(self.children) == 0:
            # Leaf
            for i in self.cluster:
                if colors[i] == 0:
                    self.reds.append(i)
                else:
                    self.blues.append(i)
        else:
            # Not a leaf
            for child in self.children:
                child.populate_colors(colors)
                self.reds.extend(child.reds)
                self.blues.extend(child.blues)


from ._vanilla import VanillaFairletDecomposition


class ScalableFairletDecomposition(VanillaFairletDecomposition):
    def __init__(self, p, q):
        self.fairlets = []
        self.fairlet_centers = []
        assert p <= q, "Please use balance parameters in the correct order"
        self.p = p
        self.q = q
        self.sensgroups = SensitiveGroups()

    def decompose(self, node, dataset, donelist, depth):
        p = self.p
        q = self.q

        # Leaf
        if len(node.children) == 0:
            node.reds = [i for i in node.reds if donelist[i] == 0]
            node.blues = [i for i in node.blues if donelist[i] == 0]
            assert self.balanced(
                p, q, len(node.reds), len(node.blues)
            ), "Reached unbalanced leaf"
            return super().decompose(node.blues, node.reds, dataset)

        # Preprocess children nodes to get rid of points that have already been clustered
        for child in node.children:
            child.reds = [i for i in child.reds if donelist[i] == 0]
            child.blues = [i for i in child.blues if donelist[i] == 0]

        R = [len(child.reds) for child in node.children]
        B = [len(child.blues) for child in node.children]

        if sum(R) == 0 or sum(B) == 0:
            assert (
                sum(R) == 0 and sum(B) == 0
            ), "One color class became empty for this node while the other did not"
            return 0

        NR = 0
        NB = 0

        # Phase 1: Add must-remove nodes
        for i in range(len(node.children)):
            if R[i] >= B[i]:
                must_remove_red = max(0, R[i] - int(np.floor(B[i] * q * 1.0 / p)))
                R[i] -= must_remove_red
                NR += must_remove_red
            else:
                must_remove_blue = max(0, B[i] - int(np.floor(R[i] * q * 1.0 / p)))
                B[i] -= must_remove_blue
                NB += must_remove_blue

        # Calculate how many points need to be added to smaller class until balance
        if NR >= NB:
            # Number of missing blues in (NR,NB)
            missing = max(0, int(np.ceil(NR * p * 1.0 / q)) - NB)
        else:
            # Number of missing reds in (NR,NB)
            missing = max(0, int(np.ceil(NB * p * 1.0 / q)) - NR)

        # Phase 2: Add may-remove nodes until (NR,NB) is balanced or until no more such nodes
        for i in range(len(node.children)):
            if missing == 0:
                assert self.balanced(p, q, NR, NB), "Something went wrong"
                break
            if NR >= NB:
                may_remove_blue = B[i] - int(np.ceil(R[i] * p * 1.0 / q))
                remove_blue = min(may_remove_blue, missing)
                B[i] -= remove_blue
                NB += remove_blue
                missing -= remove_blue
            else:
                may_remove_red = R[i] - int(np.ceil(B[i] * p * 1.0 / q))
                remove_red = min(may_remove_red, missing)
                R[i] -= remove_red
                NR += remove_red
                missing -= remove_red

        # Phase 3: Add unsatuated fairlets until (NR,NB) is balanced
        for i in range(len(node.children)):
            if self.balanced(p, q, NR, NB):
                break
            if R[i] >= B[i]:
                num_saturated_fairlets = int(R[i] / q)
                excess_red = R[i] - q * num_saturated_fairlets
                excess_blue = B[i] - p * num_saturated_fairlets
            else:
                num_saturated_fairlets = int(B[i] / q)
                excess_red = R[i] - p * num_saturated_fairlets
                excess_blue = B[i] - q * num_saturated_fairlets
            R[i] -= excess_red
            NR += excess_red
            B[i] -= excess_blue
            NB += excess_blue

        assert self.balanced(p, q, NR, NB), "Constructed node sets are unbalanced"

        reds = []
        blues = []
        for i in range(len(node.children)):
            for j in node.children[i].reds[R[i] :]:
                reds.append(j)
                donelist[j] = 1
            for j in node.children[i].blues[B[i] :]:
                blues.append(j)
                donelist[j] = 1

        assert len(reds) == NR and len(blues) == NB, "Something went horribly wrong"
        return super().decompose(blues, reds, dataset) + sum(
            [
                self.decompose(child, dataset, donelist, depth + 1)
                for child in node.children
            ]
        )

    def fit_transform(self, dataset, group_a, group_b):
        sensitive_groups = np.c_[group_a, group_b]
        colors = self.sensgroups.fit_transform(sensitive_groups, convert_numeric=True)
        root = build_quadtree(dataset)
        "Main fairlet clustering function, returns cost wrt original metric (not tree metric)"
        p = self.p
        q = self.q
        root.populate_colors(colors)
        assert self.balanced(
            p, q, len(root.reds), len(root.blues)
        ), "Dataset is unbalanced"
        root.populate_colors(colors)
        donelist = [0] * dataset.shape[0]
        cost = self.decompose(root, dataset, donelist, 0)
        return self.fairlets, self.fairlet_centers, cost


### QUADTREE CODE ###


def build_quadtree(dataset, max_levels=0, random_shift=True):
    "If max_levels=0 there no level limit, quadtree will partition until all clusters are singletons"
    dimension = dataset.shape[1]
    lower = np.amin(dataset, axis=0)
    upper = np.amax(dataset, axis=0)

    shift = np.zeros(dimension)
    if random_shift:
        for d in range(dimension):
            spread = upper[d] - lower[d]
            shift[d] = np.random.uniform(0, spread)
            upper[d] += spread

    return build_quadtree_aux(
        dataset, range(dataset.shape[0]), lower, upper, max_levels, shift
    )


def build_quadtree_aux(dataset, cluster, lower, upper, max_levels, shift):
    """
    "lower" is the "bottom-left" (in all dimensions) corner of current hypercube
    "upper" is the "upper-right" (in all dimensions) corner of current hypercube
    """

    dimension = dataset.shape[1]
    cell_too_small = True
    for i in range(dimension):
        if upper[i] - lower[i] > EPSILON:
            cell_too_small = False

    node = TreeNode()
    if max_levels == 1 or len(cluster) <= 1 or cell_too_small:
        # Leaf
        node.set_cluster(cluster)
        return node

    # Non-leaf
    midpoint = 0.5 * (lower + upper)
    subclusters = defaultdict(list)
    for i in cluster:
        subclusters[
            tuple([dataset[i, d] + shift[d] <= midpoint[d] for d in range(dimension)])
        ].append(i)
    for edge, subcluster in subclusters.items():
        sub_lower = np.zeros(dimension)
        sub_upper = np.zeros(dimension)
        for d in range(dimension):
            if edge[d]:
                sub_lower[d] = lower[d]
                sub_upper[d] = midpoint[d]
            else:
                sub_lower[d] = midpoint[d]
                sub_upper[d] = upper[d]
        node.add_child(
            build_quadtree_aux(
                dataset, subcluster, sub_lower, sub_upper, max_levels - 1, shift
            )
        )
    return node



import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator

from holisticai.utils.transformers.bias import BMInprocessing as BMImp
from holisticai.utils.transformers.bias import SensitiveGroups

from .algorithm import FairScoreClassifierAlgorithm


class FairScoreClassifier(BaseEstimator, BMImp):
    """
    Generates a classification model that integrates fairness constraints for multiclass classification. This algorithm
    returns a matrix of lambda coefficients that scores a given input vector. The higher the score, the higher the probability
    of the input vector to be classified as the majority class.

    References:
        Julien Rouzot, Julien Ferry, Marie-José Huguet. Learning Optimal Fair Scoring Systems for Multi-
        Class Classification. ICTAI 2022 - The 34th IEEE International Conference on Tools with Artificial
        Intelligence, Oct 2022, Virtual, United States. 
    """

    def __init__(
        self,
        objectives: dict,
        constraints: dict = {},
        lambda_bound: int = 9,
        time_limit: int = 100,
    ):
        """
        Init FairScoreClassifier object

        Parameters
        ----------
        objectives : dict
            The weighted objectives list to be optimized.

        constraints : dict
            The constraints list to be used in the optimization. The keys are the constraints names and the values are the bounds.

        lambda_bound : int
            Lower and upper bound for the scoring system cofficients.
        """
        self.objectives = objectives
        self.constraints = constraints
        self.lambda_bound = lambda_bound
        self.time_limit = time_limit

    def fit(self, X, y, group_a, group_b):
        """
        Fit model using Grid Search Algorithm.

        Parameters
        ----------

        X : matrix-like
            input matrix

        y : numpy array
            target vector

        protected_groups : list
            The sensitive groups.

        protected_labels : list
            The senstive labels.

        sensitive_features : numpy array
            Matrix where each columns is a sensitive feature e.g. [col_1=group_a, col_2=group_b]

        Returns
        -------
        the same object
        """
        sensgroup = SensitiveGroups()
        p_attr = sensgroup.fit_transform(
            np.stack([group_a, group_b], axis=1), convert_numeric=True
        )
        Xtrain = np.hstack([np.ones((X.shape[0], 1)), X, p_attr.values.reshape(-1, 1)])
        fairness_groups = [Xtrain.shape[1] - 1]
        fairness_labels = np.arange(y.shape[1]).tolist()

        self.model_ = FairScoreClassifierAlgorithm(
            self.objectives,
            fairness_groups,
            fairness_labels,
            self.constraints,
            self.lambda_bound,
            self.time_limit,
        )
        self.model_.fit(Xtrain, y)
        return self

    def predict(self, X, group_a, group_b):
        sensgroup = SensitiveGroups()
        p_attr = sensgroup.fit_transform(
            np.stack([group_a, group_b], axis=1), convert_numeric=True
        )
        X_ = np.hstack([np.ones((X.shape[0], 1)), X, p_attr.values.reshape(-1, 1)])
        preds = self.model_.predict(X_)
        y_pred = pd.DataFrame(preds, columns=np.arange(preds.shape[1]))
        return y_pred.idxmax(axis=1).values

    def transform_estimator(self, estimator):
        self.estimator = estimator
        return self



from typing import Optional, Union

import numpy as np
from sklearn.base import BaseEstimator

from holisticai.bias.mitigation.commons.fairlet_clustering.decompositions import (
    DecompositionMixin,
    ScalableFairletDecomposition,
    VanillaFairletDecomposition,
)
from holisticai.bias.mitigation.inprocessing.fairlet_clustering.algorithm import (
    FairletClusteringAlgorithm,
)
from holisticai.utils.models.cluster import KCenters, KMedoids
from holisticai.utils.transformers.bias import BMInprocessing as BMImp

DECOMPOSITION_CATALOG = {
    "Scalable": ScalableFairletDecomposition,
    "Vanilla": VanillaFairletDecomposition,
}
CLUSTERING_CATALOG = {"KCenters": KCenters, "KMedoids": KMedoids}


class FairletClustering(BaseEstimator, BMImp):
    """
    Fairlet Clustering inprocessing bias mitigation works in two steps:
    - The pointset is partitioned into subsets called fairlets that satisfy
    the fairness requirement and approximately preserve the k-median objective.
    - Fairlets are merged into k clusters by one of the existing k-median algorithms.

    Reference
    ---------
        Backurs, Arturs, et al. "Scalable fair clustering." International Conference on
        Machine Learning. PMLR, 2019.
    """

    def __init__(
        self,
        n_clusters: Optional[int],
        decomposition: Union["str", "DecompositionMixin"] = "Vanilla",
        clustering_model: Optional["str"] = "KCenters",
        p: Optional[str] = 1,
        q: Optional[float] = 3,
        seed: Optional[int] = None,
    ):
        """
        Parameters
        ----------
            n_clusters : int
                The number of clusters to form as well as the number of centroids to generate.

            decomposition : str
                Fairlet decomposition strategy, available: Vanilla, Scalable

            clustering_model : str
                specified lambda parameter

            p : int
                fairlet decomposition parameter for Vanilla and Scalable strategy

            q : int
                fairlet decomposition parameter for Vanilla and Scalable strategy

            seed : int
                Random seed.
        """
        if decomposition in ["Scalable", "Vanilla"]:
            self.decomposition = DECOMPOSITION_CATALOG[decomposition](p=p, q=q)
        self.clustering_model = CLUSTERING_CATALOG[clustering_model](
            n_clusters=n_clusters
        )

        # Constant parameters
        self.algorithm = FairletClusteringAlgorithm(
            decomposition=self.decomposition, clustering_model=self.clustering_model
        )
        self.p = p
        self.q = q
        self.n_clusters = n_clusters
        self.seed = seed

    def fit(
        self,
        X: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """
        Fit the model

        Description
        -----------
        Learn a fair cluster.

        Parameters
        ----------

        X : numpy array
            input matrix

        group_a : numpy array
            binary mask vector

        group_b : numpy array
            binary mask vector

        Returns
        -------
        the same object
        """
        params = self._load_data(X=X, group_a=group_a, group_b=group_b)
        X = params["X"]
        group_a = params["group_a"].astype("int32")
        group_b = params["group_b"].astype("int32")
        np.random.seed(self.seed)
        self.algorithm.fit(X, group_a=group_a, group_b=group_b)
        return self

    @property
    def cluster_centers_(self):
        return self.algorithm.cluster_centers_

    @property
    def labels_(self):
        return self.algorithm.labels

    def predict(self, X: np.ndarray):
        """
        Prediction

        Description
        ----------
        Predict cluster for the given samples.

        Parameters
        ----------
        X : pandas.DataFrame or numpy array
            Test samples.

        Returns
        -------

        numpy.ndarray: Predicted output per sample.
        """
        params = self._load_data(X=X)
        X = params["X"]
        return self.algorithm.predict(X)

    def fit_predict(self, X: np.ndarray, group_a: np.ndarray, group_b: np.ndarray):
        """
        Prediction

        Description
        ----------
        Fit and Predict the cluster for the given samples.

        Parameters
        ----------
        X : pandas.DataFrame or numpy array
            Test samples.

        group_a : numpy array
            binary mask vector

        group_b : numpy array
            binary mask vector


        Returns
        -------

        numpy.ndarray: Predicted cluster per sample.
        """
        self.fit(X, group_a, group_b)
        return self.labels_



from typing import Optional

import numpy as np
from tqdm import tqdm

from holisticai.utils.models.recommender._rsbase import RecommenderSystemBase
from holisticai.utils.transformers.bias import BMInprocessing as BMImp

from .common_utils.propensity_utils import popularity_model_propensity
from .common_utils.utils import updateP, updateQ


class PopularityPropensityMF(BMImp, RecommenderSystemBase):
    """
    Popularity Propensity Matrix Factorization can be used for Recommender Systems.
    This model is trained with propensity matrix factorization defined in (Eq. 1) [1]. Here the propensity P(u,i)
    is estimated based on popularity.

    References:
        [1] Tobias Schnabel, Adith Swaminathan, Ashudeep Singh, Navin Chandak, and
        Thorsten Joachims. 2016. Recommendations as treatments: Debiasing learning
        and evaluation. arXiv preprint arXiv:1602.05352 (2016).
    """

    def __init__(
        self,
        K: Optional[int] = 10,
        beta: Optional[float] = 0.02,
        steps: Optional[int] = 100,
        verbose: Optional[int] = 0,
    ):
        """
        Init Popularity Propensity Matrix Factorization

        Parameters
        ----------

        K : int
            Specifies the number of dimensions.

        beta : float
            Parameter used to update P and Q.

        steps : int
            Number of iterations.

        verbose : int
            If >0, will show progress percentage.
        """
        self.K = K
        self.beta = beta
        self.steps = steps
        self.verbose = verbose

    def fit(self, X: Optional[np.ndarray], **kargs):
        """
        Fit model

        Parameters
        ----------

        X : matrix-like
            rating matrix, 0 means non-raked cases.

        P0 : matrix-like (optional)
            Initial P matrix (numUsers, K)

        Q0 : matrix-like (optional)
            Initial P matrix (numItems, K)
        """
        numUsers, numItems = X.shape
        P0 = kargs.get("P", np.random.rand(numUsers, self.K))
        Q0 = kargs.get("Q", np.random.rand(numItems, self.K))
        propensity = popularity_model_propensity(X)
        P, Q = self._coordinate_descent_matrix_factorization(X, P0, Q0, propensity)
        self.pred = np.dot(P, Q)
        self.invP = None
        return self

    def _coordinate_descent_matrix_factorization(self, R, P, Q, propensity):
        """
        Coordinate Descent Matrix Factorization Algorithm
        """
        N, M = R.shape
        Q = Q.T
        K = P.shape[1]
        b = self.beta * np.identity(K)
        invprop = 1 / propensity
        u_rated_items = [np.where(R[u, :] > 0)[0] for u in range(N)]
        i_rated_users = [np.where(R[:, i] > 0)[0] for i in range(M)]
        for _ in tqdm(range(self.steps), leave=self.verbose > 0):
            P = updateP(P, b, R, Q, u_rated_items, invprop)
            Q = updateQ(P, b, R, Q, i_rated_users, invprop)
        return P, Q



# imports
from .exponentiated_gradient.transformer import ExponentiatedGradientReduction
from .fair_k_center_clustering.transformer import FairKCenterClustering
from .fair_k_mediam_clustering.transformer import FairKmedianClustering
from .fair_scoring_classifier.transformer import FairScoreClassifier
from .fairlet_clustering.transformer import FairletClustering
from .grid_search.transformer import GridSearchReduction
from .matrix_factorization.blind_spot_aware import BlindSpotAwareMF
from .matrix_factorization.debiasing_learning.transformer import DebiasingLearningMF
from .matrix_factorization.popularity_propensity import PopularityPropensityMF
from .meta_fair_classifier.transformer import MetaFairClassifier
from .prejudice_remover.transformer import PrejudiceRemover
from .two_sided_fairness.transformer import FairRec
from .variational_fair_clustering.transformer import VariationalFairClustering

__all__ = [
    "ExponentiatedGradientReduction",
    "GridSearchReduction",
    "PrejudiceRemover",
    "MetaFairClassifier",
    "VariationalFairClustering",
    "FairKCenterClustering",
    "FairKmedianClustering",
    "FairletClustering",
    "BlindSpotAwareMF",
    "DebiasingLearningMF",
    "PopularityPropensityMF",
    "FairRec",
    "FairScoreClassifier",
]


import importlib

torch_spec = importlib.util.find_spec("torch")
if torch_spec is not None:
    from .adversarial_debiasing.torch.transformer import AdversarialDebiasing

__all__ += ["AdversarialDebiasing"]



from typing import Optional

import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin, clone

from holisticai.utils.transformers.bias import BMInprocessing as BMImp

from ..commons.classification import _constraints as cc
from ..commons.regression import _constraints as rc
from ..commons.regression import _losses as rl
from .algorithm import ExponentiatedGradientAlgorithm


class ExponentiatedGradientReduction(BaseEstimator, ClassifierMixin, BMImp):
    """
    Exponentiated gradient reduction is an in-processing technique that reduces
    fair classification to a sequence of cost-sensitive classification problems,
    returning a randomized classifier with the lowest empirical error subject to
    fair classification constraints.

    References
    ----------
        Agarwal, Alekh, et al. "A reductions approach to fair classification."
        International Conference on Machine Learning. PMLR, 2018.
    """

    CONSTRAINTS = [
        "DemographicParity",
        "EqualizedOdds",
        "TruePositiveRateParity",
        "FalsePositiveRateParity",
        "ErrorRateParity",
    ]

    def __init__(
        self,
        constraints: str = "EqualizedOdds",
        eps: Optional[float] = 0.01,
        max_iter: Optional[int] = 50,
        nu: Optional[float] = None,
        eta0: Optional[float] = 2.0,
        loss: str = "ZeroOne",
        min_val: float = None,
        max_val: float = None,
        upper_bound: float = 0.01,
        verbose: Optional[int] = 0,
        estimator=None,
    ):

        """

        Parameters
        ----------
        estimator : sklearn-like
            The model you want to mitigate bias for.

        constraints (str or BaseMoment): If string, keyword
            denoting the :class:`BaseMoment` object
            defining the disparity constraints:
            [
                "DemographicParity",
                "EqualizedOdds",
                "TruePositiveRateParity",
                "FalsePositiveRateParity",
                "ErrorRateParity",
            ]

        eps: Allowed fairness constraint violation; the solution is
            guaranteed to have the error within ``2*best_gap`` of the best
            error under constraint eps; the constraint violation is at most
            ``2*(eps+best_gap)``.

        T: Maximum number of iterations.

        nu: Convergence threshold for the duality gap, corresponding to a
            conservative automatic setting based on the statistical
            uncertainty in measuring classification error.

        eta_mul: Initial setting of the learning rate.

        drop_prot_attr: Boolean flag indicating whether to drop protected
            attributes from training data.

        loss : str
            String identifying loss function for constraints. Options include "ZeroOne", "Square", and "Absolute."

        min_val : float
            Loss function parameter for "Square" and "Absolute," typically the minimum of the range of y values.

        max_val: float
            Loss function parameter for "Square" and "Absolute," typically the maximum of the range of y values.

        verbose : int
            If >0, will show progress percentage.
        """

        self.constraints = constraints
        self.eps = eps
        self.max_iter = max_iter
        self.nu = nu
        self.eta0 = eta0
        self.loss = loss
        self.min_val = min_val
        self.max_val = max_val
        self.upper_bound = upper_bound
        self.verbose = verbose
        self.estimator = estimator

    def transform_estimator(self, estimator):

        """
        This method is deprecated but retained for backwards-compatibility. You should pass the estimator object directly in the constructor.
        """

        self.estimator = estimator
        return self

    def fit(
        self,
        X: np.ndarray,
        y_true: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """
        Fit Exponentiated Gradient Reduction

        Parameters
        ----------
        X : matrix-like
            Input matrix
        y_true : array-like
            Target vector
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Returns
        -------
        Self

        """
        params = self._load_data(X=X, y_true=y_true, group_a=group_a, group_b=group_b)
        group_a = params["group_a"]
        group_b = params["group_b"]
        X = params["X"]
        y_true = params["y_true"]

        sensitive_features = np.stack([group_a, group_b], axis=1)

        self.estimator_ = clone(self.estimator)

        constraints_catalog = {
            "DemographicParity": cc.DemographicParity,
            "EqualizedOdds": cc.EqualizedOdds,
            "TruePositiveRateParity": cc.TruePositiveRateParity,
            "FalsePositiveRateParity": cc.FalsePositiveRateParity,
            "ErrorRateParity": cc.ErrorRateParity,
            "BoundedGroupLoss": rc.BoundedGroupLoss,
        }

        constraint_kargs = self._constraint_parameters()
        self.constraint_ = constraints_catalog[self.constraints](**constraint_kargs)

        self.model_ = ExponentiatedGradientAlgorithm(
            self.estimator_,
            constraints=self.constraint_,
            eps=self.eps,
            max_iter=self.max_iter,
            nu=self.nu,
            eta0=self.eta0,
            verbose=self.verbose,
        )

        self.model_.fit(X, y_true, sensitive_features=sensitive_features)

        return self

    def _constraint_parameters(self):
        kargs = {}
        if self.constraints == "BoundedGroupLoss":
            losses = {
                "ZeroOne": rl.ZeroOneLoss,
                "Square": rl.SquareLoss,
                "Absolute": rl.AbsoluteLoss,
            }
            if self.loss == "ZeroOne":
                self.loss_ = losses[self.loss]()
            else:
                self.loss_ = losses[self.loss](self.min_val, self.max_val)
            kargs.update({"loss": self.loss_, "upper_bound": self.upper_bound})
        return kargs

    def predict(self, X):
        """
        Prediction

        Description
        ----------
        Predict output for the given samples.

        Parameters
        ----------
        X : matrix-like
            Input matrix.

        Returns
        -------
        numpy.ndarray: Predicted output
        """
        return self.model_.predict(X).ravel()

    def predict_proba(self, X):
        """
        Predict Probabilities

        Description
        ----------
        Probability estimate for the given samples.

        Parameters
        ----------
        X : matrix-like
            Input matrix.

        Returns
        -------
        numpy.ndarray: probability output
        """
        return self.model_.predict_proba(X)
